# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'revpiinfo.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_diag_revpiinfo(object):
    def setupUi(self, diag_revpiinfo):
        diag_revpiinfo.setObjectName("diag_revpiinfo")
        diag_revpiinfo.resize(627, 398)
        self.gridLayout = QtWidgets.QGridLayout(diag_revpiinfo)
        self.gridLayout.setObjectName("gridLayout")
        self.lbl_head = QtWidgets.QLabel(diag_revpiinfo)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.lbl_head.setFont(font)
        self.lbl_head.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_head.setObjectName("lbl_head")
        self.gridLayout.addWidget(self.lbl_head, 0, 0, 1, 3)
        self.lbl_lbl_version_pyload = QtWidgets.QLabel(diag_revpiinfo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_lbl_version_pyload.setFont(font)
        self.lbl_lbl_version_pyload.setObjectName("lbl_lbl_version_pyload")
        self.gridLayout.addWidget(self.lbl_lbl_version_pyload, 3, 0, 1, 1)
        self.lst_files = QtWidgets.QListWidget(diag_revpiinfo)
        self.lst_files.setObjectName("lst_files")
        self.gridLayout.addWidget(self.lst_files, 3, 2, 4, 1)
        self.lbl_version_pyload = QtWidgets.QLabel(diag_revpiinfo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_version_pyload.setFont(font)
        self.lbl_version_pyload.setText("0.0.0")
        self.lbl_version_pyload.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_version_pyload.setObjectName("lbl_version_pyload")
        self.gridLayout.addWidget(self.lbl_version_pyload, 3, 1, 1, 1)
        self.btn_box = QtWidgets.QDialogButtonBox(diag_revpiinfo)
        self.btn_box.setOrientation(QtCore.Qt.Horizontal)
        self.btn_box.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.btn_box.setObjectName("btn_box")
        self.gridLayout.addWidget(self.btn_box, 7, 0, 1, 3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lbl_lbl_version_control = QtWidgets.QLabel(diag_revpiinfo)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.lbl_lbl_version_control.setFont(font)
        self.lbl_lbl_version_control.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_lbl_version_control.setObjectName("lbl_lbl_version_control")
        self.horizontalLayout.addWidget(self.lbl_lbl_version_control)
        self.lbl_version_control = QtWidgets.QLabel(diag_revpiinfo)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.lbl_version_control.setFont(font)
        self.lbl_version_control.setText("0.0.0")
        self.lbl_version_control.setObjectName("lbl_version_control")
        self.horizontalLayout.addWidget(self.lbl_version_control)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 3)
        self.lbl_info = QtWidgets.QLabel(diag_revpiinfo)
        self.lbl_info.setMinimumSize(QtCore.QSize(300, 0))
        self.lbl_info.setWordWrap(True)
        self.lbl_info.setObjectName("lbl_info")
        self.gridLayout.addWidget(self.lbl_info, 4, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 5, 0, 1, 1)
        self.lbl_link = QtWidgets.QLabel(diag_revpiinfo)
        self.lbl_link.setText("<html><head/><body><p><a href=\"https://revpimodio.org/\"><span style=\" text-decoration: underline; color:#0000ff;\">https://revpimodio.org/</span></a></p></body></html>")
        self.lbl_link.setOpenExternalLinks(True)
        self.lbl_link.setObjectName("lbl_link")
        self.gridLayout.addWidget(self.lbl_link, 6, 0, 1, 2)

        self.retranslateUi(diag_revpiinfo)
        self.btn_box.accepted.connect(diag_revpiinfo.accept) # type: ignore
        self.btn_box.rejected.connect(diag_revpiinfo.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(diag_revpiinfo)

    def retranslateUi(self, diag_revpiinfo):
        _translate = QtCore.QCoreApplication.translate
        diag_revpiinfo.setWindowTitle(_translate("diag_revpiinfo", "Program information"))
        self.lbl_head.setText(_translate("diag_revpiinfo", "RevPi Python PLC - Commander"))
        self.lbl_lbl_version_pyload.setText(_translate("diag_revpiinfo", "RevPiPyLoad version on RevPi:"))
        self.lbl_lbl_version_control.setText(_translate("diag_revpiinfo", "Version:"))
        self.lbl_info.setText(_translate("diag_revpiinfo", "RevPiModIO, RevPiPyLoad and RevPiPyControl are community driven projects. They are all free and open source software.\n"
"All of them comes with ABSOLUTELY NO WARRANTY, to the extent permitted by\n"
"applicable law.\n"
"\n"
"(c) Sven Sager, License: GPLv2"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    diag_revpiinfo = QtWidgets.QDialog()
    ui = Ui_diag_revpiinfo()
    ui.setupUi(diag_revpiinfo)
    diag_revpiinfo.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""Package: RevPiCommander."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"
__package__ = "revpicommander"
__version__ = "0.10.0"



# -*- coding: utf-8 -*-
"""Setup script for RevPiCommander."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2018 Sven Sager"
__license__ = "GPLv2"

from setuptools import find_namespace_packages, setup

from src.revpicommander import __version__

setup(
    name="revpicommander",
    version=__version__,

    packages=find_namespace_packages("src"),
    package_dir={'': 'src'},
    include_package_data=True,

    install_requires=[
        "keyring",
        "PyQt5",
        "paramiko",
        "revpimodio2",
        "zeroconf"
    ],
    entry_points={
        'gui_scripts': [
            'revpicommander = revpicommander.revpicommander:main',
        ],
    },

    platforms=["all"],

    url="https://revpimodio.org/revpipyplc/",
    license="GPLv2",
    author="Sven Sager",
    author_email="akira@narux.de",
    maintainer="Sven Sager",
    maintainer_email="akira@revpimodio.org",
    description="GUI for Revolution Pi to upload programs and do IO-Checks",
    long_description="The RevPiCommander is a GUI tool to manage your Revolution Pi over the\n"
                     "network. You can search for RevPis in your network, manage the settings\n"
                     "of RevPiPyLoad and do IO checks on your local machine. Developing your\n"
                     "control program is very easy with the developer, upload and debug it\n"
                     "over the network.",
    keywords=["revpi", "revolution pi", "revpimodio", "plc"],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ],
)



# -*- coding: utf-8 -*-
"""One device of the Revolution Pi."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import struct

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from . import proginit as pi
from .ui.debugios_ui import Ui_win_debugios


class DebugIos(QtWidgets.QMainWindow, Ui_win_debugios):
    """IO window of one device."""

    device_closed = QtCore.pyqtSignal(int)
    """This window was closed."""

    search_class = (QtWidgets.QLineEdit, QtWidgets.QDoubleSpinBox, QtWidgets.QCheckBox)

    def __init__(self, position: int, name: str, inputs: list, outputs: list, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.restoreGeometry(helper.cm.settings.debug_geos.get(position, b''))
        self.setWindowTitle("{0} - {1}".format(position, name))
        self.gb_io.setTitle(self.gb_io.title().format(name))

        self.__qwa = {}
        """Quick widget access."""

        self.position = position
        self.name = name
        self.inputs = inputs.copy()
        self.outputs = outputs.copy()
        self.write_values = False

        self.length = self._calc_device_length(self.inputs, self.outputs)

        self.style_sheet = ""
        self._create_io(self.inputs, self.saw_inp, True)
        self._create_io(self.outputs, self.saw_out, False)
        self.style_sheet = "background-color: red;"

    def __del__(self):
        pi.logger.debug("DebugIos.__del__")

    def closeEvent(self, a0: QtGui.QCloseEvent):
        pi.logger.debug("DebugIos.closeEvent")
        helper.cm.settings.debug_geos[self.position] = self.saveGeometry()
        self.device_closed.emit(self.position)

    @staticmethod
    def _calc_min_max(byte_length: int, signed: bool):
        """Calculate min an max value which fits to bytes."""
        max_int_value = 256 ** byte_length
        return max_int_value / 2 * -1 if signed else 0.0, \
            max_int_value / 2 - 1 if signed else max_int_value - 1

    @staticmethod
    def _calc_device_length(inputs: list, outputs: list) -> int:
        """Calculate the device length with IO data."""
        if inputs and outputs:
            min_input = min(inputs, key=lambda k: k[2])
            max_output = max(outputs, key=lambda k: k[2])
        elif inputs:
            min_input = min(inputs, key=lambda k: k[2])
            max_output = max(inputs, key=lambda k: k[2])
        elif outputs:
            min_input = min(outputs, key=lambda k: k[2])
            max_output = max(outputs, key=lambda k: k[2])
        else:
            return 0

        return max_output[2] + max_output[1] - min_input[2]

    def _create_io(self, lst_ios: list, container: QtWidgets.QWidget, read_only: bool):
        lst_names = list(lst[0] for lst in lst_ios)
        layout = container.layout()  # type: QtWidgets.QFormLayout

        for val in container.findChildren(self.search_class, options=QtCore.Qt.FindDirectChildrenOnly):
            name = val.objectName()
            if name not in lst_names:
                # Remove old io from layout
                del self.__qwa[name]
                layout.removeRow(layout.getWidgetPosition(val)[0])

        counter = -1
        for io in lst_ios:
            counter += 1

            name = io[0]
            byte_length = io[1]
            bit_address = io[4]
            byteorder = io[5]
            signed = io[6]
            word_order = io[7] if len(io) > 7 else "ignored"

            val = container.findChild(self.search_class, name)
            if val is not None:
                # Destroy IO if the properties was changed
                if byte_length != val.property("byte_length") or \
                        bit_address != val.property("bit_address") or \
                        byteorder != ("big" if val.property("big_endian") else "little") or \
                        signed != val.property("signed"):
                    del self.__qwa[name]
                    layout.removeRow(layout.getWidgetPosition(val)[0])
                    pi.logger.debug("Destroy property changed IO '{0}'".format(name))
                else:
                    continue

            lbl = QtWidgets.QLabel(name, container)
            lbl.setObjectName("lbl_".format(name))
            lbl.setStyleSheet(self.style_sheet)

            val = self._create_widget(name, byte_length, bit_address, byteorder, signed, read_only, word_order)
            val.setParent(container)
            layout.insertRow(counter, val, lbl)

        self.splitter.setSizes([1, 1])

    def _create_widget(
            self, name: str, byte_length: int, bit_address: int, byteorder: str, signed: bool, read_only: bool,
            word_order: str):
        """Create widget in functions address space to use lambda functions."""
        if bit_address >= 0:
            val = QtWidgets.QCheckBox()
            val.setEnabled(not read_only)

            # Set alias to use the same function name on all widget types
            val.setValue = val.setChecked
            if not read_only:
                val.stateChanged.connect(self._change_cbx_value)
                val.value = val.isChecked

        elif byte_length > 4:
            # Bytes or string
            val = QtWidgets.QLineEdit()
            val.setReadOnly(read_only)
            val.setProperty("struct_type", "text")

            val.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            val.customContextMenuRequested.connect(self.on_context_menu)

            # Set alias to use the same function name on all widget types
            val.setValue = val.setText
            if not read_only:
                val.textChanged.connect(self._change_txt_value)
                val.value = val.text

        else:
            struct_type = "B" if byte_length == 1 else "H" if byte_length == 2 else "I"

            val = QtWidgets.QDoubleSpinBox()
            val.setReadOnly(read_only)
            val.setProperty("struct_type", struct_type)
            val.setProperty("frm", "{0}{1}".format(
                ">" if byteorder == "big" else "<",
                struct_type.lower() if signed else struct_type
            ))

            val.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            val.customContextMenuRequested.connect(self.on_context_menu)

            val.setDecimals(0)
            min_value, max_value = self._calc_min_max(byte_length, signed)
            val.setMinimum(min_value)
            val.setMaximum(max_value)
            if not read_only:
                val.valueChanged.connect(self._change_sbx_dvalue)

        val.setObjectName(name)
        val.setProperty("big_endian", byteorder == "big")
        val.setProperty("bit_address", bit_address)
        val.setProperty("byte_length", byte_length)
        val.setProperty("signed", signed)
        val.setProperty("word_order", word_order)

        self.__qwa[name] = val
        return val

    @QtCore.pyqtSlot(int)
    def _change_cbx_value(self, value: int):
        """Change value of a check box."""
        if self.sender().property("last_value") == (value == 2):
            self.sender().setStyleSheet("")
        else:
            self.sender().setStyleSheet("background-color: yellow;")

    @QtCore.pyqtSlot(float)
    def _change_sbx_dvalue(self, value: float):
        """Change value of a spin box."""
        if self.sender().property("last_value") == int(value):
            self.sender().setStyleSheet("")
        else:
            self.sender().setStyleSheet("background-color: yellow;")

    @QtCore.pyqtSlot(str)
    def _change_txt_value(self, value: str):
        """Change value of a text box."""
        if self.sender().property("last_value") == value:
            self.sender().setStyleSheet("")
        else:
            self.sender().setStyleSheet("background-color: yellow;")

    @QtCore.pyqtSlot(QtCore.QPoint)
    def on_context_menu(self, point: QtCore.QPoint):
        """Generate menu for data format changes."""
        pi.logger.debug("DebugIos.on_context_menu")

        sender = self.sender()
        men = QtWidgets.QMenu(sender)

        if sender.property("byte_length") > 4:
            # Textbox needs format buttons
            act_as_text = QtWidgets.QAction(self.tr("as text"))
            men.addAction(act_as_text)
            act_as_number = QtWidgets.QAction(self.tr("as number"))
            men.addAction(act_as_number)
            men.addSeparator()
        else:
            act_as_text = None
            act_as_number = None

        act_signed = QtWidgets.QAction(self.tr("signed"), men)
        act_signed.setCheckable(True)
        act_signed.setChecked(sender.property("signed") or False)
        men.addAction(act_signed)

        act_byteorder = QtWidgets.QAction(self.tr("big_endian"), men)
        act_byteorder.setCheckable(True)
        act_byteorder.setChecked(sender.property("big_endian") or False)
        men.addAction(act_byteorder)

        if sender.property("byte_length") > 2:
            act_wordorder = QtWidgets.QAction(self.tr("switch wordorder"))
            act_wordorder.setCheckable(True)
            act_wordorder.setChecked(sender.property("word_order") == "big")
            men.addAction(act_wordorder)
        else:
            act_wordorder = None

        rc = men.exec(sender.mapToGlobal(point))
        if not rc:
            men.deleteLater()
            return

        # Get actual value to reformat it
        actual_value, last_value = self.get_value(sender.objectName())
        if rc == act_signed:
            sender.setProperty("signed", act_signed.isChecked())
            if type(sender) == QtWidgets.QDoubleSpinBox:
                # Recalculate min / max for spinbox
                min_value, max_value = self._calc_min_max(
                    sender.property("byte_length"), sender.property("signed")
                )
                sender.setMinimum(min_value)
                sender.setMaximum(max_value)
        elif rc == act_byteorder:
            sender.setProperty("big_endian", act_byteorder.isChecked())
        elif rc == act_wordorder:
            sender.setProperty("word_order", "big" if act_wordorder.isChecked() else "little")

        if sender.property("frm"):
            sender.setProperty("frm", "{0}{1}".format(
                ">" if act_byteorder.isChecked() else "<",
                sender.property("struct_type").lower() if act_signed.isChecked()
                else sender.property("struct_type").upper()
            ))
        elif rc == act_as_text:
            sender.setProperty("struct_type", "text")
        elif rc == act_as_number:
            sender.setProperty("struct_type", "number")

        self.set_value(sender.objectName(), actual_value)
        men.deleteLater()

    def reset_label_colors(self):
        """Clean up color from labels."""
        for wid in self.findChildren(QtWidgets.QLabel):  # type: QtWidgets.QLabel
            wid.setStyleSheet("")

    def reset_change_value_colors(self, io_name=None):
        """
        Clean up color from changed outputs.

        :param io_name: Clean up only this IO
        """
        pi.logger.debug("DebugIos.reset_change_value_colors")
        if io_name is None:
            lst_wid = self.saw_out.findChildren(
                self.search_class, options=QtCore.Qt.FindDirectChildrenOnly)
        else:
            lst_wid = self.saw_out.findChildren(
                self.search_class, name=io_name, options=QtCore.Qt.FindDirectChildrenOnly)

        for wid in lst_wid:
            value, last_value = self.get_value(wid.objectName())
            if value == last_value:
                wid.setStyleSheet("")
            else:
                wid.setStyleSheet("background-color: yellow;")

    def update_ios(self, inputs: list, outputs: list):
        """Update IOs after driver reset of piCtory."""

        # Check device length, this has to match to reuse this device
        if self.length != self._calc_device_length(inputs, outputs):
            return False

        # Remove IOs, which was remove or renamed
        self._create_io(inputs, self.saw_inp, True)
        self._create_io(outputs, self.saw_out, False)

        return True

    def get_value(self, io_name: str):
        """
        Standard get function for a value of different widgets and last value.

        :param io_name: Name of IO
        :return: (actual value, last value) <class 'tuple'>
        """
        # child = self.findChild(self.search_class, io_name)
        child = self.__qwa[io_name]
        actual_value = child.value()
        last_value = child.value() if child.property("last_value") is None else child.property("last_value")
        if child.property("frm"):
            return struct.pack(child.property("frm"), int(actual_value)), \
                struct.pack(child.property("frm"), int(last_value))
        elif type(actual_value) == str:
            if child.property("struct_type") == "number":
                try:
                    actual_value = int(actual_value).to_bytes(
                        child.property("byte_length"),
                        byteorder="big" if child.property("big_endian") else "little",
                        signed=child.property("signed") or False
                    )
                    last_value = int(last_value).to_bytes(
                        child.property("byte_length"),
                        byteorder="big" if child.property("big_endian") else "little",
                        signed=child.property("signed") or False
                    )
                    return actual_value, last_value
                except Exception:
                    pi.logger.error("Could not convert '{0}' to bytes".format(actual_value))
                    pass

            return actual_value.encode(), last_value.encode()
        else:
            return actual_value, last_value

    def set_value(self, io_name: str, value, just_last_value=False):
        """
        Standard set function for a value of different widgets.

        :param io_name: Name of IO
        :param value: New value as bytes or bool for widget
        :param just_last_value: Just set last value property
        """
        child = self.__qwa[io_name]

        if child.property("word_order") == "big" and type(value) == bytes:
            value = helper.swap_word_order(value)

        if child.property("frm"):
            value = struct.unpack(child.property("frm"), value)[0]
        elif type(value) == bytes:
            if child.property("struct_type") == "text":
                try:
                    value = value.decode("utf-8")
                except UnicodeDecodeError:
                    child.setProperty("struct_type", "number")
                    QtWidgets.QMessageBox.warning(
                        self, self.tr("Can not use format text"), self.tr(
                            "Can not convert bytes {0} to a text for IO '{1}'. Switch to number format instead!"
                        ).format(value, io_name)
                    )
            if child.property("struct_type") == "number":
                # fixme: Crashs with too much bytes
                value = str(int.from_bytes(
                    value,
                    byteorder="big" if child.property("big_endian") else "little",
                    signed=child.property("signed") or False
                ))

        child.setProperty("last_value", value)
        if not just_last_value:
            child.setValue(value)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'sshauth.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_diag_sshauth(object):
    def setupUi(self, diag_sshauth):
        diag_sshauth.setObjectName("diag_sshauth")
        diag_sshauth.setWindowModality(QtCore.Qt.ApplicationModal)
        diag_sshauth.resize(331, 225)
        self.verticalLayout = QtWidgets.QVBoxLayout(diag_sshauth)
        self.verticalLayout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.verticalLayout.setObjectName("verticalLayout")
        self.wid_password = QtWidgets.QWidget(diag_sshauth)
        self.wid_password.setObjectName("wid_password")
        self.formLayout = QtWidgets.QFormLayout(self.wid_password)
        self.formLayout.setObjectName("formLayout")
        self.lbl_username = QtWidgets.QLabel(self.wid_password)
        self.lbl_username.setObjectName("lbl_username")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lbl_username)
        self.lbl_password = QtWidgets.QLabel(self.wid_password)
        self.lbl_password.setObjectName("lbl_password")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lbl_password)
        self.txt_password = QtWidgets.QLineEdit(self.wid_password)
        self.txt_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.txt_password.setObjectName("txt_password")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.txt_password)
        self.txt_username = QtWidgets.QLineEdit(self.wid_password)
        self.txt_username.setObjectName("txt_username")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.txt_username)
        self.verticalLayout.addWidget(self.wid_password)
        self.cbx_save_password = QtWidgets.QCheckBox(diag_sshauth)
        self.cbx_save_password.setObjectName("cbx_save_password")
        self.verticalLayout.addWidget(self.cbx_save_password)
        self.btn_box = QtWidgets.QDialogButtonBox(diag_sshauth)
        self.btn_box.setOrientation(QtCore.Qt.Horizontal)
        self.btn_box.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.btn_box.setObjectName("btn_box")
        self.verticalLayout.addWidget(self.btn_box)
        self.lbl_info = QtWidgets.QLabel(diag_sshauth)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_info.sizePolicy().hasHeightForWidth())
        self.lbl_info.setSizePolicy(sizePolicy)
        self.lbl_info.setWordWrap(True)
        self.lbl_info.setObjectName("lbl_info")
        self.verticalLayout.addWidget(self.lbl_info)
        self.verticalLayout.setStretch(3, 1)

        self.retranslateUi(diag_sshauth)
        self.btn_box.accepted.connect(diag_sshauth.accept) # type: ignore
        self.btn_box.rejected.connect(diag_sshauth.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(diag_sshauth)

    def retranslateUi(self, diag_sshauth):
        _translate = QtCore.QCoreApplication.translate
        diag_sshauth.setWindowTitle(_translate("diag_sshauth", "SSH authentication"))
        self.lbl_username.setText(_translate("diag_sshauth", "SSH username:"))
        self.lbl_password.setText(_translate("diag_sshauth", "SSH password:"))
        self.cbx_save_password.setToolTip(_translate("diag_sshauth", "Username and password will be saved in secured operating systems\'s password storage."))
        self.cbx_save_password.setText(_translate("diag_sshauth", "Save username and password"))
        self.lbl_info.setText(_translate("diag_sshauth", "Note: The default user for SSH is \"pi\" which differs from the web configuration. You can find the password on the sticker on the device."))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    diag_sshauth = QtWidgets.QDialog()
    ui = Ui_diag_sshauth()
    ui.setupUi(diag_sshauth)
    diag_sshauth.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""Authentication dialog for SSH."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

from logging import getLogger

import keyring
from PyQt5 import QtWidgets
from keyring.errors import KeyringError

from .ui.sshauth_ui import Ui_diag_sshauth

log = getLogger()


class SSHAuth(QtWidgets.QDialog, Ui_diag_sshauth):

    def __init__(self, user_name="", service_name: str = None, parent=None):
        """
        Ask the user for username and password or use saved entries.

        If you want to use the operating system's password storage, you have
        to set a 'service_name'. The value must be unique for your application
        or for each user, if the username is the same.

        :param user_name: Preset username, also used to check password save
        :param service_name: Identity to save passwords in os's password save
        :param parent: Qt parent for this dialog
        """
        log.debug("SSHAuth.__init__")

        super().__init__(parent)
        self.setupUi(self)

        self._in_keyring = False
        self._service_name = service_name
        self.cbx_save_password.setVisible(bool(service_name))
        self.txt_username.setText(user_name)

    def accept(self) -> None:
        log.debug("SSHAuth.accept")

        if self._service_name and self.cbx_save_password.isChecked():
            try:
                keyring.set_password(self._service_name, self.username, self.password)
            except KeyringError as e:
                log.error(e)
                self._in_keyring = False
                QtWidgets.QMessageBox.warning(
                    self, self.tr("Could not save password"), self.tr(
                        "Could not save password to operating systems password save.\n\n"
                        "Maybe your operating system does not support saving passwords. "
                        "This could be due to missing libraries or programs.\n\n"
                        "This is not an error of RevPi Commander."
                    )
                )
            else:
                self._in_keyring = True

        super().accept()

    def exec(self) -> int:
        log.debug("SSHAuth.exec")

        if self._service_name:
            try:
                saved_password = keyring.get_password(self._service_name, self.username)
            except KeyringError as e:
                log.error(e)
                self._in_keyring = False
            else:
                if saved_password:
                    self._in_keyring = True
                    self.txt_password.setText(saved_password)
                    return QtWidgets.QDialog.Accepted

        return super().exec()

    def remove_saved_password(self) -> None:
        """Remove saved password."""
        log.debug("SSHAuth.remove_saved_password")

        if self._service_name:
            try:
                keyring.delete_password(self._service_name, self.username)
            except KeyringError as e:
                log.error(e)

    @property
    def in_keyring(self) -> bool:
        """True, if password is in keyring."""
        return self._in_keyring

    @property
    def password(self) -> str:
        """Get the saved or entered password."""
        return self.txt_password.text()

    @property
    def username(self) -> str:
        """Get the entered username."""
        return self.txt_username.text()



# -*- coding: utf-8 -*-
"""File manager for up und download PLC program."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

import gzip
import os
from enum import IntEnum
from xmlrpc.client import Binary

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from . import proginit as pi
from .backgroundworker import BackgroundWorker
from .helper import WidgetData
from .ui.files_ui import Ui_win_files


class NodeType(IntEnum):
    FILE = 1000
    DIR = 1001


class UploadFiles(BackgroundWorker):

    def __init__(self, file_list: list, parent):
        super().__init__(parent)
        self.ec = 1
        self.file_list = file_list
        self.plc_program_included = False  # Will be True, when opt_program was found in files

    def run(self) -> None:
        self.steps_todo.emit(len(self.file_list))

        # Get config to find actual auto start program for warnings
        opt_program = helper.cm.call_remote_function("get_config", default_value={})
        opt_program = opt_program.get("plcprogram", "none.py")

        progress_counter = 0
        for file_name in self.file_list:
            progress_counter += 1

            # Remove base dir of file to set relative for PyLoad
            send_name = file_name.replace(helper.cm.settings.watch_path, "")[1:]
            self.status_message.emit(send_name)

            # Check whether this is the auto start program
            if send_name == opt_program:
                self.plc_program_included = True

            # Transfer file
            try:
                with open(file_name, "rb") as fh:
                    upload_status = helper.cm.call_remote_function(
                        "plcupload", Binary(gzip.compress(fh.read())), send_name,
                        default_value=False
                    )
            except Exception as e:
                pi.logger.error(e)
                self.ec = -2
                return

            if not upload_status:
                self.ec = -1
                return

            self.steps_done.emit(progress_counter)
            if self.check_cancel():
                return

        self.ec = 0


class RevPiFiles(QtWidgets.QMainWindow, Ui_win_files):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.dc_settings = {}
        self.tree_files_counter = 0
        self.tree_files_counter_max = 10000
        self.lbl_path_local.setText(helper.cm.settings.watch_path or self.tr("Please select..."))
        self.lbl_path_local.setToolTip(self.lbl_path_local.text())

        self.btn_all.setEnabled(False)
        self.btn_to_left.setEnabled(False)
        self.btn_to_right.setEnabled(False)
        self.btn_delete_revpi.setEnabled(False)

        if helper.cm.settings.watch_path:
            self._load_files_local(True)
        if helper.cm.connected:
            self._load_files_revpi(True)

        self.restoreGeometry(helper.settings.value("files/geo", b''))
        self.splitter.setSizes(list(map(int, helper.settings.value("files/splitter", [0, 0]))))

    def __del__(self):
        pi.logger.debug("RevPiFiles.__del__")

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        pi.logger.debug("RevPiFiles.closeEvent")
        helper.settings.setValue("files/geo", self.saveGeometry())
        helper.settings.setValue("files/splitter", self.splitter.sizes())

    def _do_my_job(self, stop_restart=True):
        """
        Upload the selected files and do a optionally restart.

        :param stop_restart: True will restart program
        """
        if not helper.cm.connected:
            return

        if stop_restart and helper.cm.call_remote_function("plcstop") is None:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not stop plc program on Revolution Pi."
                )
            )
            return

        uploader = UploadFiles(self.file_list_local(), self)
        if uploader.exec_dialog(self.tr("File transfer...")) == QtWidgets.QDialog.Rejected:
            return

        if uploader.ec == 0:
            # Tell user, we did not find the auto start program in files
            if not uploader.plc_program_included:
                QtWidgets.QMessageBox.information(
                    self, self.tr("Information"), self.tr(
                        "A PLC program has been uploaded. Please check the "
                        "PLC program settings to see if the correct program "
                        "is specified as the start program."
                    )
                )

        elif uploader.ec == -1:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "The Revolution Pi could not process some parts of the "
                    "transmission."
                )
            )

        elif uploader.ec == -2:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"),
                self.tr("Errors occurred during transmission")
            )

        if stop_restart and helper.cm.call_remote_function("plcstart", default_value=1) != 0:
            QtWidgets.QMessageBox.warning(
                self, self.tr("Warning"), self.tr(
                    "Could not start the plc program on Revolution Pi."
                )
            )

    def _set_gui_control_states(self):
        """Setup states of actions and buttons."""
        state_local = len(self.tree_files_local.selectedItems()) > 0
        state_revpi = len(self.tree_files_revpi.selectedItems()) > 0

        self.btn_all.setEnabled(state_local)
        self.btn_to_right.setEnabled(state_local)

        if "plcdeletefile" not in helper.cm.xml_funcs:
            self.btn_delete_revpi.setEnabled(False)
            self.btn_delete_revpi.setToolTip(self.tr("The RevPiPyLoad version on the Revolution Pi is to old."))
        else:
            self.btn_delete_revpi.setEnabled(state_revpi)
        if "plcdownload_file" not in helper.cm.xml_funcs:
            self.btn_to_left.setEnabled(False)
            self.btn_to_left.setToolTip(self.tr("The RevPiPyLoad version on the Revolution Pi is to old."))
        elif not helper.cm.settings.watch_path:
            self.btn_to_left.setEnabled(False)
            self.btn_to_left.setToolTip(self.tr("Choose a local directory first."))
        else:
            self.btn_to_left.setEnabled(state_revpi)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: Tree management

    @staticmethod
    def _parent_selection_state(item: QtWidgets.QTreeWidgetItem):
        """Set selected, if all children are selected."""
        if item.parent():
            all_selected = True
            for i in range(item.parent().childCount()):
                if not item.parent().child(i).isSelected():
                    all_selected = False
                    break
            item.parent().setSelected(all_selected)

    def _select_children(self, top_item: QtWidgets.QTreeWidgetItem, value: bool):
        """Recursive select children from parent."""
        pi.logger.debug("RevPiFiles._select_children")

        for i in range(top_item.childCount()):
            item = top_item.child(i)
            if item.type() == NodeType.DIR:
                item.setSelected(value)
                self._select_children(item, value)
            elif item.type() == NodeType.FILE:
                item.setSelected(value)

    def __item_selection_changed(self, tree_view: QtWidgets.QTreeWidget):
        """Manager vor item selection of three views."""
        item = tree_view.currentItem()
        if item is None:
            return

        pi.logger.debug("RevPiFiles.__itemSelectionChanged")

        # Block while preselect other entries
        tree_view.blockSignals(True)

        if item.type() == NodeType.DIR:
            self._select_children(item, item.isSelected())
        elif item.type() == NodeType.FILE:
            self._parent_selection_state(item)

        tree_view.blockSignals(False)

        self._set_gui_control_states()

    @QtCore.pyqtSlot()
    def on_tree_files_local_itemSelectionChanged(self):
        self.__item_selection_changed(self.tree_files_local)
        helper.cm.settings.watch_files = self.file_list_local()

    @QtCore.pyqtSlot()
    def on_tree_files_revpi_itemSelectionChanged(self):
        self.__item_selection_changed(self.tree_files_revpi)

    # endregion # # # # #

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: Local file lists

    def __insert_files_local(self, base_dir: str, child=None):
        """
        Recursively add files to tree view.

        :param base_dir: Directory to scan for files
        :param child: Child widget to add new widgets
        """
        if not os.path.exists(base_dir):
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not open last directory '{0}'."
                ).format(base_dir)
            )
            return

        for de in os.scandir(base_dir):  # type: os.DirEntry

            if self.tree_files_counter > self.tree_files_counter_max:
                return

            if de.is_dir(follow_symlinks=False):
                item = QtWidgets.QTreeWidgetItem(NodeType.DIR)
                item.setText(0, de.name)
                item.setIcon(0, QtGui.QIcon(":/main/ico/folder.ico"))
                if child:
                    child.addChild(item)
                else:
                    self.tree_files_local.addTopLevelItem(item)

                self.__insert_files_local(de.path, item)

            elif de.is_file(follow_symlinks=False):
                self.tree_files_counter += 1

                item = QtWidgets.QTreeWidgetItem(NodeType.FILE)
                item.setText(0, de.name)
                item.setData(0, WidgetData.file_name, de.path)
                item.setIcon(0, QtGui.QIcon(
                    ":/file/ico/file-else.ico" if de.name.find(".py") == -1 else
                    ":/file/ico/file-python.ico"
                ))
                if child:
                    child.addChild(item)
                else:
                    self.tree_files_local.addTopLevelItem(item)

                item.setSelected(de.path in helper.cm.settings.watch_files)
                self._parent_selection_state(item)

    def _load_files_local(self, silent=False):
        """
        Refresh the file list.

        :param silent: Do not show message boxes
        """
        pi.logger.debug("RevPiFiles._load_files_local")

        self.tree_files_counter = 0
        self.tree_files_local.blockSignals(True)
        self.tree_files_local.clear()
        self.__insert_files_local(helper.cm.settings.watch_path)
        self.tree_files_local.sortItems(0, QtCore.Qt.AscendingOrder)
        self.tree_files_local.blockSignals(False)

        if not silent and self.tree_files_counter > self.tree_files_counter_max:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Stop scanning for files, because we found more than {0} files."
                ).format(self.tree_files_counter_max)
            )

        self._set_gui_control_states()

    def file_list_local(self):
        """Generate a file list with full path of selected entries."""
        pi.logger.debug("RevPiFiles.file_list_local")
        lst = []
        for item in self.tree_files_local.selectedItems():
            if item.type() == NodeType.DIR:
                # We just want files
                continue
            lst.append(item.data(0, WidgetData.file_name))

        return lst

    # endregion # # # # #

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # region #      REGION: RevPi file lists

    def _load_files_revpi(self, silent=False):
        """
        Refresh the file list of revolution pi.

        :param silent: Do not show message boxes
        """
        pi.logger.debug("RevPiFiles._load_files_revpi")

        self.tree_files_revpi.blockSignals(True)
        self.tree_files_revpi.clear()
        self.tree_files_revpi.blockSignals(False)

        if not helper.cm.connected:
            lst_revpi = None
        else:
            lst_revpi = helper.cm.call_remote_function("get_filelist")
            # Just load settings once
            if not self.dc_settings:
                self.dc_settings = helper.cm.call_remote_function("get_config", default_value={})
                self.lbl_path_revpi.setText(
                    self.dc_settings.get("plcworkdir", self.tr("Could not load path of working dir"))
                )
                self.lbl_path_revpi.setToolTip(self.lbl_path_revpi.text())

        if lst_revpi is not None:
            lst_revpi.sort()

            for path_file in lst_revpi:
                lst_path_file = path_file.split("/")
                dir_node = None  # type: QtWidgets.QTreeWidgetItem

                for folder in lst_path_file[:-1]:
                    new_dir_node = QtWidgets.QTreeWidgetItem(NodeType.DIR)
                    new_dir_node.setText(0, folder)
                    new_dir_node.setIcon(0, QtGui.QIcon(":/main/ico/folder.ico"))

                    if dir_node:
                        # Subfolder of top level
                        for i in range(dir_node.childCount()):
                            item = dir_node.child(i)
                            if item.type() != NodeType.DIR:
                                continue
                            if item.text(0) == new_dir_node.text(0):
                                dir_node = item
                                new_dir_node = None
                                break
                        if new_dir_node:
                            dir_node.addChild(new_dir_node)
                            dir_node = new_dir_node
                    else:
                        # Search in top level
                        for i in range(self.tree_files_revpi.topLevelItemCount()):
                            item = self.tree_files_revpi.topLevelItem(i)
                            if item.type() != NodeType.DIR:
                                continue
                            if item.text(0) == new_dir_node.text(0):
                                dir_node = item
                                new_dir_node = None
                                break
                        if new_dir_node:
                            self.tree_files_revpi.addTopLevelItem(new_dir_node)
                            dir_node = new_dir_node

                # This is the file name
                object_name = lst_path_file[-1]
                item = QtWidgets.QTreeWidgetItem(NodeType.FILE)
                item.setText(0, object_name)
                item.setData(0, WidgetData.file_name, path_file)
                item.setIcon(0, QtGui.QIcon(
                    ":/file/ico/file-else.ico" if object_name.find(".py") == -1 else
                    ":/file/ico/file-python.ico"
                ))
                if dir_node:
                    dir_node.addChild(item)
                else:
                    self.tree_files_revpi.addTopLevelItem(item)

            self.tree_files_revpi.sortItems(0, QtCore.Qt.AscendingOrder)

        elif not silent:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not load file list from Revolution Pi."
                )
            )

        self._set_gui_control_states()

    def file_list_revpi(self):
        """Generate a file list with full path of selected entries."""
        pi.logger.debug("RevPiFiles.file_list_revpi")
        lst = []
        for item in self.tree_files_revpi.selectedItems():
            if item.type() == NodeType.DIR:
                # We just want files
                continue
            lst.append(item.data(0, WidgetData.file_name))

        return lst

    # endregion # # # # #

    @QtCore.pyqtSlot()
    def on_btn_all_clicked(self):
        pi.logger.debug("RevPiFiles.on_btn_all_clicked")
        self._do_my_job(True)
        self.file_list_revpi()

    @QtCore.pyqtSlot()
    def on_btn_select_local_clicked(self):
        pi.logger.debug("RevPiFiles.on_btn_select_clicked")

        diag_folder = QtWidgets.QFileDialog(
            self, self.tr("Select folder..."),
            helper.cm.settings.watch_path,
        )
        diag_folder.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
        if diag_folder.exec() != QtWidgets.QFileDialog.Accepted:
            return

        selected_dir = diag_folder.selectedFiles()[0]

        if not os.access(selected_dir, os.R_OK):
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "Can not access the folder '{0}' to read files."
                )
            )
            helper.cm.settings.watch_files = []
            helper.cm.settings.watch_path = ""
            return

        self.lbl_path_local.setText(selected_dir)
        self.lbl_path_local.setToolTip(self.lbl_path_local.text())
        helper.cm.settings.watch_path = selected_dir
        helper.cm.settings.watch_files = []

        self._load_files_local(False)

    @QtCore.pyqtSlot()
    def on_btn_refresh_local_clicked(self):
        pi.logger.debug("RevPiFiles.on_btn_refresh_clicked")
        self._load_files_local(False)

    @QtCore.pyqtSlot()
    def on_btn_refresh_revpi_clicked(self):
        pi.logger.debug("RevPiFiles.on_btn_refresh_revpi_clicked")
        self._load_files_revpi(False)

    @QtCore.pyqtSlot()
    def on_btn_to_right_clicked(self):
        """Upload selected files to revolution pi."""
        pi.logger.debug("RevPiFiles.on_btn_to_right_clicked")
        self._do_my_job(False)
        self._load_files_revpi(True)

    @QtCore.pyqtSlot()
    def on_btn_to_left_clicked(self):
        """Download selected file."""
        pi.logger.debug("RevPiFiles.on_btn_to_left_clicked")

        override = None
        for item in self.tree_files_revpi.selectedItems():
            if item.type() != NodeType.FILE:
                continue

            file_name = item.data(0, WidgetData.file_name)
            rc = helper.cm.call_remote_function(
                "plcdownload_file", file_name,
                default_value=Binary()
            )
            rc = rc.data
            if not rc:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error..."), self.tr(
                        "Error while download file '{0}'."
                    ).format(file_name)
                )
            else:
                file_name = os.path.join(helper.cm.settings.watch_path, file_name)
                if override is None and os.path.exists(file_name):
                    rc_diag = QtWidgets.QMessageBox.question(
                        self, self.tr("Override files..."), self.tr(
                            "One or more files does exist on your computer! Do you want to override the existing"
                            "files?\n\nSelect 'Yes' to override, 'No' to download only missing files."
                        ),
                        buttons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel
                    )
                    if rc_diag == QtWidgets.QMessageBox.Cancel:
                        return
                    override = rc_diag == QtWidgets.QMessageBox.Yes

                if os.path.exists(file_name) and not override:
                    pi.logger.debug("Skip existing file '{0}'".format(file_name))
                    continue

                os.makedirs(os.path.dirname(file_name), exist_ok=True)
                file_data = gzip.decompress(rc)
                with open(os.path.join(helper.cm.settings.watch_path, file_name), "wb") as fh:
                    fh.write(file_data)

        self._load_files_local()

    @QtCore.pyqtSlot()
    def on_btn_delete_revpi_clicked(self):
        """Remove selected files from working directory on revolution pi."""
        pi.logger.debug("RevPiFiles.btn_delete_revpi_clicked")

        lst_delete = []
        for item in self.tree_files_revpi.selectedItems():
            if item.type() == NodeType.FILE:
                lst_delete.append(item.data(0, WidgetData.file_name))

        rc = QtWidgets.QMessageBox.question(
            self, self.tr("Delete files from Revolution Pi..."), self.tr(
                "Do you want to delete {0} files from revolution pi?"
            ).format(len(lst_delete))
        )
        if rc != QtWidgets.QMessageBox.Yes:
            return

        for file_name in lst_delete:
            rc = helper.cm.call_remote_function("plcdeletefile", file_name, default_value=False)
            if not rc:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Error..."), self.tr(
                        "Error while delete file '{0}'."
                    ).format(file_name)
                )

        self._load_files_revpi()



# -*- coding: utf-8 -*-
"""RevPiPyLoad options window."""
__author__ = "Sven Sager"
__copyright__ = "Copyright (C) 2023 Sven Sager"
__license__ = "GPLv2"

from PyQt5 import QtCore, QtGui, QtWidgets

from . import helper
from . import proginit as pi
from .aclmanager import AclManager
from .mqttmanager import MqttManager
from .ui.revpioption_ui import Ui_diag_options


class RevPiOption(QtWidgets.QDialog, Ui_diag_options):
    """Set options of RevPiPyLoad."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.setFixedSize(self.size())

        self.dc = {}
        self.acl_plcserver = ""
        self.acl_xmlrpc = ""
        self.mrk_xml_ask = True
        self.wrong_names = False

        self._dict_mqttsettings = {
            "mqttbasetopic": "revpi01",
            "mqttclient_id": "",
            "mqttbroker_address": "127.0.0.1",
            "mqttpassword": "",
            "mqttport": 1883,
            "mqttsend_on_event": 0,
            "mqttsendinterval": 30,
            "mqtttls_set": 0,
            "mqttusername": "",
            "mqttwrite_outputs": 0,
        }

        self.diag_aclmanager = AclManager(self)
        self.rejected.connect(self.diag_aclmanager.reject)
        self.diag_mqttmanager = MqttManager(self)
        self.rejected.connect(self.diag_mqttmanager.reject)
        self.diag_mqttmanager.dc = self._dict_mqttsettings

    def _apply_acl(self):
        """Set availability of controls depending on ACL level."""
        allow = helper.cm.xml_mode >= 4

        self.cbx_autostart.setEnabled(allow)
        self.cbx_autoreload.setEnabled(allow)
        self.sbx_autoreloaddelay.setEnabled(allow)
        self.cbx_zeroonexit.setEnabled(allow)
        self.cbx_zeroonerror.setEnabled(allow)
        self.cbb_replace_io.setEnabled(allow)
        self.txt_replace_io.setEnabled(allow and self.cbb_replace_io.currentIndex() == 3)
        self.cbx_plcserver.setEnabled(allow)
        self.cbx_mqtt.setEnabled(allow)
        self.cbx_xmlrpc.setEnabled(allow)

        if allow:
            self.btn_box.setStandardButtons(
                QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
            )
        else:
            self.btn_box.setStandardButtons(
                QtWidgets.QDialogButtonBox.Cancel
            )

    def _changesdone(self):
        """
        Check for unsaved changes in dialog.

        :return: True, if unsaved changes was found
        """
        return (
            int(self.cbx_autostart.isChecked()) != self.dc.get("autostart", 0) or
            int(self.cbx_autoreload.isChecked()) != self.dc.get("autoreload", 1) or
            self.sbx_autoreloaddelay.value() != self.dc.get("autoreloaddelay", 5) or
            int(self.cbx_zeroonexit.isChecked()) != self.dc.get("zeroonexit", 0) or
            int(self.cbx_zeroonerror.isChecked()) != self.dc.get("zeroonerror", 0) or
            self.txt_replace_io.text() != self.dc.get("replace_ios", "") or
            self.cbb_reset_driver_action.currentIndex() != self.dc.get("reset_driver_action", 2) or
            # todo: self.dc.get("rtlevel", 2)

            int(self.cbx_plcserver.isChecked()) != self.dc.get("plcserver", 0) or
            self.acl_plcserver != self.dc.get("plcserveracl", "") or

            int(self.cbx_mqtt.isChecked()) != self.dc.get("mqtt", 0) or
            self._changesdone_mqtt() or

            int(self.cbx_xmlrpc.isChecked()) != self.dc.get("xmlrpc", 0) or
            self.acl_xmlrpc != self.dc.get("xmlrpcacl", "")
        )

    def _changesdone_mqtt(self):
        """
        Check for unsaved changes in mqtt settings.

        :return: True, if unsaved changes was found
        """
        for key in self._dict_mqttsettings:
            if key in self.dc:
                if self._dict_mqttsettings[key] != self.dc[key]:
                    return True
        return False

    def _load_settings(self):
        """Load values to GUI widgets."""
        pi.logger.debug("RevPiOption._load_settings")

        self.mrk_xml_ask = True

        self.cbx_autostart.setChecked(bool(self.dc.get("autostart", 0)))
        self.cbx_autoreload.setChecked(bool(self.dc.get("autoreload", 1)))
        self.sbx_autoreloaddelay.setValue(self.dc.get("autoreloaddelay", 5))
        self.cbx_zeroonexit.setChecked(bool(self.dc.get("zeroonexit", 0)))
        self.cbx_zeroonerror.setChecked(bool(self.dc.get("zeroonerror", 0)))
        self.txt_replace_io.setText(self.dc.get("replace_ios", ""))
        self.cbb_reset_driver_action.setCurrentIndex(self.dc.get("reset_driver_action", 2))
        self.cbx_plcserver.setChecked(bool(self.dc.get("plcserver", 0)))
        self.acl_plcserver = self.dc.get("plcserveracl", "")
        self.cbx_mqtt.setChecked(bool(self.dc.get("mqtt", 0)))
        self.cbx_xmlrpc.setChecked(bool(self.dc.get("xmlrpc", 0)))
        self.acl_xmlrpc = self.dc.get("xmlrpcacl", "")

        # Find the right index of combo box
        if self.txt_replace_io.text() == "":
            self.cbb_replace_io.setCurrentIndex(0)
        elif self.txt_replace_io.text() == "/etc/revpipyload/replace_ios.conf":
            self.cbb_replace_io.setCurrentIndex(1)
        elif self.txt_replace_io.text() == "replace_ios.conf":
            self.cbb_replace_io.setCurrentIndex(2)
        else:
            self.cbb_replace_io.setCurrentIndex(3)

        # Update directory with mqtt values to compare with dc values
        for key in self._dict_mqttsettings:
            if key in self.dc:
                self._dict_mqttsettings[key] = self.dc[key]

    def _translate_wrong_names(self) -> dict:
        """
        Translate settings values of revpipyload < 0.10.0.

        With RevPiPyLoad 0.10.0 we replaced the words master-slave with
        client-server. Unfortunately we cannot expect that everyone will be
        able to switch to the new version of RevPiPyLoad immediately.
        Therefore, for a few versions of this software, we need to do a
        translation of the values.

        This function will translate the self.dc always to the new values and
        return a copy of it with new or old values, depending on previous
        detections.

        :return: Settings with wrong values, if detected in previous calls
        """
        name_mappings = (
            ("plcslave", "plcserver"),
            ("plcslaveacl", "plcserveracl"),
        )
        for wrong, right in name_mappings:
            if wrong in self.dc:
                self.wrong_names = True
                self.dc[right] = self.dc[wrong]
                del self.dc[wrong]

        translated_settings = self.dc.copy()
        if self.wrong_names:
            for wrong, right in name_mappings:
                if right in translated_settings:
                    translated_settings[wrong] = self.dc[right]
                    del translated_settings[right]

        return translated_settings

    def accept(self) -> None:
        if not self._changesdone():
            super().accept()
            return

        ask = QtWidgets.QMessageBox.question(
            self, self.tr("Question"), self.tr(
                "The settings will be set on the Revolution Pi now.\n\n"
                "ACL changes and service settings are applied immediately."
            )
        ) == QtWidgets.QMessageBox.Yes

        if not ask:
            return

        self.dc["autostart"] = int(self.cbx_autostart.isChecked())
        self.dc["autoreload"] = int(self.cbx_autoreload.isChecked())
        self.dc["autoreloaddelay"] = self.sbx_autoreloaddelay.value()
        self.dc["reset_driver_action"] = self.cbb_reset_driver_action.currentIndex()
        self.dc["zeroonexit"] = int(self.cbx_zeroonexit.isChecked())
        self.dc["zeroonerror"] = int(self.cbx_zeroonerror.isChecked())
        self.dc["replace_ios"] = self.txt_replace_io.text()

        # PLCServer Settings
        self.dc["plcserver"] = int(self.cbx_plcserver.isChecked())
        self.dc["plcserveracl"] = self.acl_plcserver

        # MQTT Settings
        self.dc["mqtt"] = int(self.cbx_mqtt.isChecked())
        self.dc["mqttbasetopic"] = self._dict_mqttsettings["mqttbasetopic"]
        self.dc["mqttsendinterval"] = int(self._dict_mqttsettings["mqttsendinterval"])
        self.dc["mqttsend_on_event"] = int(self._dict_mqttsettings["mqttsend_on_event"])
        self.dc["mqttwrite_outputs"] = int(self._dict_mqttsettings["mqttwrite_outputs"])
        self.dc["mqttbroker_address"] = self._dict_mqttsettings["mqttbroker_address"]
        self.dc["mqttport"] = int(self._dict_mqttsettings["mqttport"])
        self.dc["mqtttls_set"] = int(self._dict_mqttsettings["mqtttls_set"])
        self.dc["mqttusername"] = self._dict_mqttsettings["mqttusername"]
        self.dc["mqttpassword"] = self._dict_mqttsettings["mqttpassword"]
        self.dc["mqttclient_id"] = self._dict_mqttsettings["mqttclient_id"]

        # XML Settings
        self.dc["xmlrpc"] = int(self.cbx_xmlrpc.isChecked())
        self.dc["xmlrpcacl"] = self.acl_xmlrpc

        saved = helper.cm.call_remote_function(
            "set_config", self._translate_wrong_names(), ask,
            default_value=False
        )

        if saved:
            super().accept()
        else:
            QtWidgets.QMessageBox.critical(
                self, self.tr("Error"), self.tr(
                    "The settings could not be saved on the Revolution Pi!\n"
                    "Try to save the values one mor time and check the log "
                    "files of RevPiPyLoad if the error rises again."
                )
            )

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        if self._changesdone():
            ask = QtWidgets.QMessageBox.question(
                self, self.tr("Question"), self.tr(
                    "Do you really want to quit? \nUnsaved changes will be lost."
                )
            ) == QtWidgets.QMessageBox.Yes

            if ask:
                self.reject()
            else:
                a0.ignore()

    def exec(self) -> int:
        # Reset class values
        if not helper.cm.connected:
            return QtWidgets.QDialog.Rejected

        self.dc = helper.cm.call_remote_function("get_config", default_value={})
        if len(self.dc) == 0:
            return QtWidgets.QDialog.Rejected

        self.wrong_names = False
        self._translate_wrong_names()

        self._load_settings()
        self._apply_acl()

        running = helper.cm.call_remote_function(
            "plcslaverunning" if self.wrong_names else "plcserverrunning",
            default_value=False
        )
        self.lbl_server_status.setText(
            self.tr("running") if running else self.tr("stopped")
        )
        self.lbl_server_status.setStyleSheet(
            "color: green" if running else "color: red"
        )

        running = helper.cm.call_remote_function("mqttrunning")
        if running is None:
            # On older versions of RevPiPyLoad MQTT is not implemented
            self.cbx_mqtt.setToolTip(self.tr(
                "The MQTT service is not available on your RevPiPyLoad version."
            ))
            self.btn_mqtt.setVisible(False)
            self.lbl_mqtt_status.setText("N/A")
            self.lbl_mqtt_status.setStyleSheet("")
        else:
            self.cbx_mqtt.setToolTip("")
            self.btn_mqtt.setVisible(True)
            self.lbl_mqtt_status.setText(
                self.tr("running") if running else self.tr("stopped")
            )
            self.lbl_mqtt_status.setStyleSheet(
                "color: green" if running else "color: red"
            )

        return super().exec()

    def reject(self) -> None:
        """Reject all sub windows and reload settings."""
        self._load_settings()
        super().reject()

    @QtCore.pyqtSlot(int)
    def on_cbb_replace_io_currentIndexChanged(self, index: int):
        """Update replace_io path in text field to compare values."""
        if index == 0:
            self.txt_replace_io.setText("")
        elif index == 1:
            self.txt_replace_io.setText("/etc/revpipyload/replace_ios.conf")
        elif index == 2:
            self.txt_replace_io.setText("replace_ios.conf")
        else:
            self.txt_replace_io.setText(self.dc.get("replace_ios", ""))

        self.txt_replace_io.setEnabled(index == 3)

    @QtCore.pyqtSlot()
    def on_btn_aclplcserver_clicked(self):
        """Start ACL manager to edit ACL entries."""
        self.diag_aclmanager.setup_acl_manager(self.acl_plcserver, {
            0: self.tr("read only"),
            1: self.tr("read and write"),
        })
        self.diag_aclmanager.read_only = helper.cm.xml_mode < 4
        if self.diag_aclmanager.exec() == QtWidgets.QDialog.Accepted:
            self.acl_plcserver = self.diag_aclmanager.get_acl()

    @QtCore.pyqtSlot()
    def on_btn_mqtt_clicked(self):
        """Open MQTT settings."""
        if not helper.cm.connected:
            return

        self.diag_mqttmanager.read_only = helper.cm.xml_mode < 4
        self.diag_mqttmanager.exec()

    @QtCore.pyqtSlot(int)
    def on_cbx_xmlrpc_stateChanged(self, state: int):
        if state == QtCore.Qt.Unchecked and self.mrk_xml_ask:
            self.mrk_xml_ask = QtWidgets.QMessageBox.question(
                self, self.tr("Question"), self.tr(
                    "Are you sure you want to deactivate the XML-RPC server? "
                    "You will NOT be able to access the Revolution Pi with "
                    "this program after saving the options!"
                )
            ) == QtWidgets.QMessageBox.No
            if self.mrk_xml_ask:
                self.cbx_xmlrpc.setCheckState(QtCore.Qt.Checked)

    @QtCore.pyqtSlot()
    def on_btn_aclxmlrpc_clicked(self):
        self.diag_aclmanager.setup_acl_manager(self.acl_xmlrpc, {
            0: self.tr("Start/Stop PLC program and read logs"),
            1: self.tr("+ read IOs in watch mode"),
            2: self.tr("+ read properties and download PLC program"),
            3: self.tr("+ upload PLC program"),
            4: self.tr("+ set properties")
        })
        self.diag_aclmanager.read_only = helper.cm.xml_mode < 4
        if self.diag_aclmanager.exec() == QtWidgets.QDialog.Accepted:
            self.acl_xmlrpc = self.diag_aclmanager.get_acl()
